// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import com.microsoft.azure.servicebus.management.ManagementClient;
import com.microsoft.azure.servicebus.management.SubscriptionDescription;
import com.microsoft.azure.servicebus.primitives.MessagingEntityNotFoundException;
import com.microsoft.azure.servicebus.primitives.ServiceBusException;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.azure.serviceBusManager.IManagementClientFactory;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.subscriber.pubsub.PullSubscription;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@ExtendWith(MockitoExtension.class)
public class PullSubscriptionTest {
    private static final String dataPartitionId = "data-partition-id";
    private static final String subscriptionId = "subscription-id";
    private static final String topicName = "topic-name";
    private static final String errorMessage = "some error message";
    private static final String RESOURCE_PROVISIONING_ERROR_MESSAGE = "Resource Conflict Occurred. Another conflicting operation may be in progress";
    private static final String MANAGEMENT_CLIENT_ERROR_MESSAGE = "Unable to create management client";
    private static final String SERVICE_BUS_ERROR_MESSAGE = "Server can not be reached";

    @Mock
    private IManagementClientFactory factory;

    @Mock
    private JaxRsDpsLog logger;

    @Mock
    private DpsHeaders dpsHeaders;

    @Mock
    private ManagementClient serviceBusManager;

    @InjectMocks
    private PullSubscription pullSubscription;

    @BeforeEach
    public void init() {
        lenient().when(dpsHeaders.getPartitionId()).thenReturn(dataPartitionId);
    }

    @Test
    public void shouldCreatePullSubscription() throws Exception {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        SubscriptionDescription subscriptionDescription = new SubscriptionDescription(topicName, subscriptionId);
        lenient().when(serviceBusManager.createSubscription(topicName, subscriptionId)).thenReturn(subscriptionDescription);

        pullSubscription.createSubscription(subscriptionId, topicName);

        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).createSubscription(topicName, subscriptionId);
        verify(logger, times(1)).debug(String.format("Pull Subscription with id %s created successfully", subscriptionId));
    }

    @Test
    public void shouldReturn500WhenPullSubscriptionCreationFailed() throws Exception {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        when(serviceBusManager.createSubscription(topicName, subscriptionId)).thenThrow(new ServiceBusException(true, SERVICE_BUS_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.createSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).createSubscription(topicName, subscriptionId);
    }

    @Test
    public void shouldThrow409WhenConflictInPullSubscriptionCreation() throws Exception {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        when(serviceBusManager.createSubscription(topicName, subscriptionId)).thenThrow(new ServiceBusException(true, RESOURCE_PROVISIONING_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.createSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(409, exception.getError().getCode());
        assertEquals("Another request is trying to create the same Pull subscription", exception.getMessage());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).createSubscription(topicName, subscriptionId);
        verify(logger, times(1)).error("Another request is trying to create the same Pull subscription");
    }

    @Test
    public void shouldThrow500NullPointerExceptionWhenPullSubscriptionCreationFailed() throws Exception {
        lenient().when(factory.getManager(dataPartitionId)).thenThrow(new AppException(500, "Null Pointer Exception", MANAGEMENT_CLIENT_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.createSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals(MANAGEMENT_CLIENT_ERROR_MESSAGE, exception.getMessage());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(0)).createSubscription(topicName, subscriptionId);
    }

    @Test
    public void shouldDeletePullSubscription() throws Exception {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        lenient().doNothing().when(serviceBusManager).deleteSubscription(topicName, subscriptionId);

        pullSubscription.deleteSubscription(subscriptionId, topicName);

        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).deleteSubscription(topicName, subscriptionId);
        verify(logger, times(1)).debug(String.format("Pull Subscription with id %s deleted successfully", subscriptionId));
    }

    @Test
    public void shouldThrowMessagingEntityNotFoundExceptionWhenDeletingPullSubscription() throws Exception {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        when(serviceBusManager.deleteSubscription(topicName, subscriptionId)).thenThrow(new MessagingEntityNotFoundException(errorMessage));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.deleteSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(404, exception.getError().getCode());
        assertEquals(String.format("Pull Subscription with id %s does not exist.", subscriptionId), exception.getError().getMessage());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).deleteSubscription(topicName, subscriptionId);
    }

    @Test
    public void shouldThrow500ExceptionWhenDeletingPullSubscription() throws ServiceBusException, InterruptedException {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        when(serviceBusManager.deleteSubscription(topicName, subscriptionId)).thenThrow(new ServiceBusException(true, SERVICE_BUS_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.deleteSubscription(subscriptionId, topicName);
        });
        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error deleting Pull subscription", exception.getError().getMessage());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).deleteSubscription(topicName, subscriptionId);
        verify(logger, times(1)).error("Deleting Pull Subscription failed with error: " + new ServiceBusException(true, SERVICE_BUS_ERROR_MESSAGE).toString());
    }

    @Test
    public void shouldThrow500NullPointerExceptionWhenDeletingPullSubscription() throws ServiceBusException, InterruptedException {
        lenient().when(factory.getManager(dataPartitionId)).thenThrow(new AppException(500, "Null Pointer Exception", MANAGEMENT_CLIENT_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.deleteSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Null Pointer Exception", exception.getError().getReason());
        assertEquals(MANAGEMENT_CLIENT_ERROR_MESSAGE, exception.getError().getMessage());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(0)).deleteSubscription(topicName, subscriptionId);
    }

    @Test
    public void shouldGetPullSubscription() throws ServiceBusException, InterruptedException {

        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        SubscriptionDescription subscriptionDescription = new SubscriptionDescription(topicName, subscriptionId);
        when(serviceBusManager.getSubscription(topicName, subscriptionId)).thenReturn(subscriptionDescription);

        pullSubscription.checkIfSubscriptionExists(subscriptionId, topicName);

        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(1)).getSubscription(topicName, subscriptionId);
        verify(logger, times(1)).debug(String.format("Pull Subscription with id %s fetched successfully", subscriptionId));
    }

    @Test
    public void shouldThrow500NullPointerExceptionWhenFetchingPullSubscription() throws ServiceBusException, InterruptedException {

        lenient().when(factory.getManager(dataPartitionId)).thenThrow(new AppException(500, "Null Pointer Exception", MANAGEMENT_CLIENT_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.checkIfSubscriptionExists(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Null Pointer Exception", exception.getError().getReason());
        assertEquals(MANAGEMENT_CLIENT_ERROR_MESSAGE, exception.getError().getMessage());
        verify(factory, times(1)).getManager(dataPartitionId);
        verify(serviceBusManager, times(0)).getSubscription(topicName, subscriptionId);
    }

    @Test
    public void shouldThrowNotFoundExceptionExceptionWhenFetchingPullSubscription() throws ServiceBusException, InterruptedException {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        when(serviceBusManager.getSubscription(topicName, subscriptionId)).thenThrow(new MessagingEntityNotFoundException());

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.checkIfSubscriptionExists(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals(String.format("Pull Subscription with id %s does not exist.", subscriptionId), exception.getError().getMessage());
        verify(serviceBusManager, times(1)).getSubscription(topicName, subscriptionId);
        verify(logger, times(1)).error(String.format("Pull Subscription with id %s does not exist.", subscriptionId));
    }

    @Test
    public void shouldThrow500ExceptionWhenFetchingPullSubscription() throws ServiceBusException, InterruptedException {
        lenient().when(factory.getManager(dataPartitionId)).thenReturn(serviceBusManager);
        when(serviceBusManager.getSubscription(topicName, subscriptionId)).thenThrow(new ServiceBusException(true, SERVICE_BUS_ERROR_MESSAGE));

        AppException exception = assertThrows(AppException.class, () -> {
            pullSubscription.checkIfSubscriptionExists(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error while Fetching Pull subscription", exception.getError().getMessage());
        verify(serviceBusManager, times(1)).getSubscription(topicName, subscriptionId);
    }
}