/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.subscriber;

import com.google.api.gax.rpc.NotFoundException;
import com.google.cloud.pubsub.v1.SubscriptionAdminClient;
import com.google.protobuf.Duration;
import com.google.pubsub.v1.ExpirationPolicy;
import com.google.pubsub.v1.ProjectSubscriptionName;
import com.google.pubsub.v1.ProjectTopicName;
import com.google.pubsub.v1.PushConfig;
import java.io.IOException;
import java.util.Objects;
import java.util.concurrent.TimeUnit;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.opengroup.osdu.register.utils.AppServiceConfig;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

@Slf4j
@Component
public class PushSubscription {

    private static final int ACK_DEADLINE_SECONDS = 60;
    private static final int MESSAGE_RETENTION_SECONDS = 432000;
    private static final int SUBSCRIPTION_EXPIRATION_SECONDS = 31540000;
    private static final int AWAIT_TERMINATION_DURATION = 20;

    @Autowired
    private AppServiceConfig serviceConfig;

    private SubscriptionAdminClient subscriptionAdminClient;

    @PostConstruct
    public void setUpSubscriptionAdminClient() throws IOException {
        try {
            this.subscriptionAdminClient = SubscriptionAdminClient.create();
        } catch (IOException e) {
            log.error("Exception occurred while creating subscription admin client", e);
            throw e;
        }
    }

    public Subscription create(Subscription subscription, TenantInfo tenant,
        String masterGcpId, String serviceIdentityEmail)
        throws IOException {
        createPushSubscription(
            tenant.getProjectId(),
            masterGcpId,
            subscription.getTopic(),
            subscription.getNotificationId(),
            serviceConfig.getRecordsChangePubsubEndpoint(),
            serviceIdentityEmail);
        return subscription;
    }

    public void delete(String servicesProjectId, String subscriptionId) throws IOException {
        try (SubscriptionAdminClient subscriptionAdminClient = SubscriptionAdminClient.create()) {
            try {
                ProjectSubscriptionName subscriptionName = ProjectSubscriptionName.of(servicesProjectId, subscriptionId);
                subscriptionAdminClient.deleteSubscription(subscriptionName);
            } catch (NotFoundException e) {
                //silently catch
            }
        }
    }

    private com.google.pubsub.v1.Subscription createPushSubscription(String tenantProjectId, String servicesProjectId, String topicId,
        String subscriptionId, String url, String serviceIdentityEmail) {
        ProjectTopicName topicName = ProjectTopicName.of(tenantProjectId, topicId);
        ProjectSubscriptionName subscriptionName = ProjectSubscriptionName.of(servicesProjectId, subscriptionId);

        PushConfig config = PushConfig.newBuilder()
            .setPushEndpoint(url)
            .setOidcToken(PushConfig.OidcToken.newBuilder().setServiceAccountEmail(serviceIdentityEmail).build())
            .build();

        return this.subscriptionAdminClient.createSubscription(com.google.pubsub.v1.Subscription.newBuilder()
            .setName(subscriptionName.toString())
            .setTopic(topicName.toString())
            .setPushConfig(config)
            .setExpirationPolicy(ExpirationPolicy.newBuilder().setTtl(
                Duration.newBuilder().setSeconds(SUBSCRIPTION_EXPIRATION_SECONDS).build()).build())
            .setMessageRetentionDuration(Duration.newBuilder().setSeconds(
                MESSAGE_RETENTION_SECONDS).build())
            .setAckDeadlineSeconds(ACK_DEADLINE_SECONDS)
            .build());
    }

    @PreDestroy
    public void shutDownSubscriptionAdminClient() throws InterruptedException {
        if (Objects.nonNull(this.subscriptionAdminClient)) {
            log.debug("Shutting down publisher on exit, await termination duration: {} seconds", AWAIT_TERMINATION_DURATION);
            subscriptionAdminClient.shutdown();
            subscriptionAdminClient.awaitTermination(AWAIT_TERMINATION_DURATION, TimeUnit.SECONDS);
        }
    }
}
